/***************************************************************************
 * cmdforge.c: Command compiler for Mike's Ambient Device.
 * Copyright (C) 2007 Michael C McTernan,
 *    Michael.McTernan.2001@cs.bris.ac.uk
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ***************************************************************************/

/**************************************************************************
 * Include Files
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <inttypes.h>
#include <ctype.h>
#include <assert.h>
#include <string.h>
#include <unistd.h>
#include <netdb.h>
#include "seq.h"

/**************************************************************************
 * Manifest Constants
 **************************************************************************/

/**************************************************************************
 * Type Definitions
 **************************************************************************/

typedef enum
{
    FALSE = 0, TRUE
}
boolean_t;

typedef struct
{
    const char *label;
    uint8_t     byteOffset;
}
CmdLabel;

/**************************************************************************
 * Variables
 **************************************************************************/

static uint32_t lineNum = 0;
static uint32_t outBytes = 0;

static CmdLabel labelData[32];
static uint32_t labelCount = 0;

/**************************************************************************
 * Macros
 **************************************************************************/

#define MIN(a,b) ((a) < (b) ? (a) : (b))

/**************************************************************************
 * Local Functions
 **************************************************************************/

static boolean_t readLine(FILE *s, char *buffer, uint16_t len)
{
    uint16_t l = 0;
    char     c;

    assert(len > 1);

    do
    {
        c = fgetc(s);
        buffer[l++] = c;
    }
    while(c != EOF && c != '\n' && l < (len - 1));

    if(l > 1) buffer[--l] = '\0';
    if(l > 1 && buffer[l - 1] == '\r') buffer[l - 1] = '\0';

    return c != EOF;
}


static char *trimdup(const char *line)
{
    char *p, *name;

    /* Copy the name, skipping the leading colon */
    p = name = strdup(line);
    if(name == NULL)
    {
        fprintf(stderr, "Error: strdup() failed.");
        exit(EXIT_FAILURE);
    }

    /* Skip to the end of the name */
    while(*p != '\0' && !isspace(*p))
    {
        p++;
    }

    /* Null terminate */
    *p = '\0';

    return name;
}


static boolean_t getLabelIdx(const char *name, uint32_t *const idx)
{
    uint32_t t;

    for(t = 0; t < labelCount; t++)
    {
        /* Try to find the label */
        if(strcmp(name, labelData[t].label) == 0)
        {
            *idx = t;
            return TRUE;
        }
    }

    return FALSE;
}

static void stripComments(char *line)
{
    char *c;

    c = strstr(line, "/*");
    if(c) *c = '\0';

    c = strstr(line, " //");
    if(c) *c = '\0';

    c = strstr(line, ";");
    if(c) *c = '\0';
}

static boolean_t isCommentLine(const char *line)
{
    switch(*line)
    {
        case '#': case ';': case '/':
            return TRUE;
        default:
            return FALSE;
    }
}

static boolean_t isBlankLine(const char *line)
{
    while(*line != '\0')
    {
        if(!isspace(*line)) return FALSE;

        line++;
    }

    return TRUE;
}

static boolean_t isLabelLine(const char *line)
{
    return (*line == ':');
}


static void output(FILE *out, const uint8_t *b, uint16_t c)
{
#if 1
    if(out != NULL)
    {
        fwrite(b, c, 1, out);
    }
#else
    uint16_t t;

    for(t = 0; t < c; t++)
    {
        fprintf(out, "%02X", b[t]);
    }
#endif

    /* Account for the written bytes */
    outBytes += c;
}


static boolean_t writeEepromHeader(FILE *out)
{
    const char *preDefLabels[] =
    {
        "pdPost",
        "pdEthUp",
        "pdEthDown",
        "pdDhcpStart",
        "pdDhcpConfigured",
        "pdAutoIpStart",
        "pdAutoIpConfigured",
        "pdStart",
    };
    char     preDefUsr[10];
    uint16_t preDefUsrCnt;
    uint8_t  b;
    uint8_t  t;
    uint32_t labelIdx;

    /* Count any extra predefined sequences */
    preDefUsrCnt = 0;
    do
    {
        sprintf(preDefUsr, "pdUsr%d", preDefUsrCnt);
        preDefUsrCnt++;
    }
    while(getLabelIdx(preDefUsr, &labelIdx));
    preDefUsrCnt--;

    /* Now can start writing the header */
    b = sizeof(preDefLabels) / sizeof(preDefLabels[0]);
    b += preDefUsrCnt;
    if(out != NULL) output(out, &b, 1);

    /* Write out the predefined labels */
    for(t = 0; t < sizeof(preDefLabels) / sizeof(preDefLabels[0]); t++)
    {
        if(!getLabelIdx(preDefLabels[t], &labelIdx))
        {
            fprintf(stderr, "Error: Predefined label '%s' missing from eeprom script.\n",
                    preDefLabels[t]);
            return FALSE;
        }
        else if((labelData[labelIdx].byteOffset & 1) != 0)
        {
            fprintf(stderr, "Error: Predefined label '%s' malaligned.\n",
                    preDefLabels[t]);
            return FALSE;
        }
        else
        {
            b = labelData[labelIdx].byteOffset / 2;
            if(out != NULL) output(out, &b, 1);
        }
    }

    /* Write any extra predefined sequences */
    for(t = 0; t < preDefUsrCnt; t++)
    {
        sprintf(preDefUsr, "pdUsr%d", t);

        getLabelIdx(preDefUsr, &labelIdx);
        if((labelData[labelIdx].byteOffset & 1) != 0)
        {
            fprintf(stderr, "Error: User defined label '%s' malaligned.\n",
                    preDefUsr);
            return FALSE;
        }

        b = labelData[labelIdx].byteOffset / 2;
        if(out != NULL) output(out, &b, 1);
    }

    /* Write the last label */
    if(!getLabelIdx("pdEOF", &labelIdx))
    {
        fprintf(stderr, "Error: Predefined label 'pdEOF' missing from eeprom script.\n");
        return FALSE;
    }
    else
    {
        b = labelData[labelIdx].byteOffset / 2;
        if(out != NULL) output(out, &b, 1);
    }

    return TRUE;
}


static boolean_t getRegisterVal(char c, uint8_t *const reg)
{
    c = toupper(c);
    if(c < 'A' || c > 'D')
    {
        fprintf(stderr, "Error: Invalid register name (A-C)\n");
        return FALSE;
    }

    *reg = c - 'A';

    return TRUE;
}

typedef boolean_t (*handler)(FILE *out, const char *line);

static boolean_t parseColour(FILE *out, const char *line)
{
    uint32_t d, r, g, b;
    char     c;

    if(sscanf(line, "%" SCNu32 " %" SCNu32 " %" SCNu32 " %" SCNu32, &d, &r, &g, &b) == 4)
    {

        if(r > 255 || g > 255 || b > 255)
        {
            fprintf(stderr, "Error: One of more colour values out of range (0-255)\n");
            return FALSE;
        }
        else if(M_SeqMsToTick(d) > 0x7f)
        {
            fprintf(stderr, "Error: Change duration out of range (0-%d)\n", 0x2f * 32);
            return FALSE;
        }
        else
        {
            uint8_t cmd[] = { M_SeqCmdRgb(d, r, g, b) };

            output(out, cmd, sizeof(cmd));

            return TRUE;
        }
    }
    else if(sscanf(line, "%c %" SCNu32 " %" SCNu32 " %" SCNu32, &c, &r, &g, &b) == 4)
    {
        uint8_t reg;

        if(r > 255 || g > 255 || b > 255)
        {
            fprintf(stderr, "Error: One of more colour values out of range (0-255)\n");
            return FALSE;
        }
        else if(!getRegisterVal(c, &reg))
        {
            return FALSE;
        }
        else
        {
            uint8_t cmd[] = { M_SeqCmdRgbRegTicks(reg, r, g, b) };

            output(out, cmd, sizeof(cmd));

            return TRUE;
        }

    }

    return FALSE;
}


static boolean_t parseStop(FILE *out, const char *line)
{
    uint8_t cmd[] = { M_SeqCmdStop() };

    output(out, cmd, sizeof(cmd));

    return TRUE;
}


static boolean_t parseJump(FILE *out, const char *line)
{
    int8_t j;

    /* Try to parse as a signed number */
    if(sscanf(line, "%" SCNi8, &j) != 1)
    {
        uint32_t labelIdx;
        char    *name;

        /* Check for a label */
        if(*line != ':')
        {
            return FALSE;
        }

        /* Get and cleanup the name */
        name = trimdup(&line[1]);

        if(!getLabelIdx(name, &labelIdx))
        {
            /* Check if this is the initial parse for byte accounting,
             *  or actually generating the byte stream.  If the first
             *  parse, forward references may not yet be known.
             */
            if(out != NULL)
            {
                fprintf(stderr, "Error: Unknown label '%s'\n", name);
                free(name);
                return FALSE;
            }
            else
            {
                j = 0;
            }
        }
        else
        {
            j = labelData[labelIdx].byteOffset - outBytes;
        }

        free(name);
    }


    uint8_t cmd[] = { M_SeqCmdJump(j) };

    output(out, cmd, sizeof(cmd));

    return TRUE;
}


static boolean_t parseDelayMs(FILE *out, const char *line)
{
    uint32_t d;

    if(sscanf(line, "%" SCNu32, &d) != 1)
    {
        return FALSE;
    }
    else if(M_SeqMsToTick(d) > 0x3f)
    {
        fprintf(stderr, "Error: Delay out of range (0-%u)\n", 0x3f * 4 * 8);
        return FALSE;
    }
    else
    {
        uint8_t cmd[] = { M_SeqCmdDelayMs(d) };

        output(out, cmd, sizeof(cmd));

        return TRUE;
    }
}


static boolean_t parseDelaySec(FILE *out, const char *line)
{
    uint32_t d;

    if(sscanf(line, "%" SCNu32, &d) != 1)
    {
        return FALSE;
    }
    else if(d > 0x0fff)
    {
        fprintf(stderr, "Error: Delay out of range (0-%u)", 0x0fff);

        return FALSE;
    }
    else
    {
        uint8_t cmd[] = { M_SeqCmdDelaySec(d) };

        output(out, cmd, sizeof(cmd));

        return TRUE;
    }
}


static boolean_t parseAck(FILE *out, const char *line)
{
    uint8_t code;

    /* Parse the code */
    if(sscanf(line, "%" SCNu8, &code) != 1)
    {
        return FALSE;
    }

    /* Construct and write command buffer */
    uint8_t cmd[] = { M_SeqCmdAckHost(code) };

    output(out, cmd, sizeof(cmd));

    return TRUE;
}


static boolean_t parseFetch(FILE *out, const char *url)
{
    struct hostent *hent;
    const char     *page;
    char            hostname[1024];

    /* Test for and skip the http:// prefix */
    if(strncmp(url, "http://", 7) != 0)
    {
        return FALSE;
    }

    url += 7;

    /* Copy the url into the hostname array */
    strncpy(hostname, url, sizeof(hostname));

    /* Check if there is a page */
    page = strstr(url, "/");
    if(page != NULL)
    {
        /* Remove the page part from the hostname */
        hostname[page - url] = '\0';
    }

    /* Lookup the IP address */
    hent = gethostbyname(hostname);
    if(hent == NULL)
    {
        fprintf(stderr, "Error: Failed to lookup '%s'\n", hostname);
        return FALSE;
    }
    else
    {
        uint8_t cmd[] = { M_SeqCmdReadWeb(hent->h_addr[0], hent->h_addr[1],
                                          hent->h_addr[2], hent->h_addr[3]) };
        uint8_t hostLen = strlen(hostname);

        /* Write out the command and hostname without terminating null */
        output(out, cmd, sizeof(cmd));
        output(out, &hostLen, 1);
        output(out, (uint8_t *)hostname, hostLen);

        /* Append the page, without terminating null  */
        if(page != NULL)
        {
            uint8_t pageLen = strlen(page);

            output(out, &pageLen, 1);
            output(out, (uint8_t *)page, pageLen);
        }
        else
        {
            uint8_t pageLen = 0;
            output(out, &pageLen, 1);
        }
    }

    return TRUE;
}


static boolean_t parseLoad(FILE *out, const char *line)
{
    char       c;
    uint32_t val;
    uint8_t  reg;

    if(sscanf(line, "%c %" SCNu32, &c, &val) != 2 ||
       !getRegisterVal(c, &reg))
    {
        return FALSE;
    }

    if(val > 255 )
    {
        fprintf(stderr, "Error: Value out of range (0-255)\n");
        return FALSE;
    }

    uint8_t cmd[] = { M_SeqCmdRegSet(reg, val) };

    output(out, cmd, sizeof(cmd));

    return TRUE;
}


static boolean_t parseAdd(FILE *out, const char *line)
{
    char       c;
    uint32_t val;
    uint8_t  reg;

    if(sscanf(line, "%c %" SCNu32, &c, &val) != 2 ||
       !getRegisterVal(c, &reg))
    {
        return FALSE;
    }

    if(val > 0x3f)
    {
        fprintf(stderr, "Error: Value out of range (0-63)\n");
        return FALSE;
    }

    uint8_t cmd[] = { M_SeqCmdAddReg(reg, val) };

    output(out, cmd, sizeof(cmd));

    return TRUE;
}


static boolean_t parseDec(FILE *out, const char *line)
{
    char       c;
    uint32_t val;
    uint8_t  reg;

    if(sscanf(line, "%c %" SCNu32, &c, &val) != 2 ||
       !getRegisterVal(c, &reg))
    {
        return FALSE;
    }

    if(val > 0x3f)
    {
        fprintf(stderr, "Error: Value out of range (0-63)\n");
        return FALSE;
    }

    uint8_t cmd[] = { M_SeqCmdDecReg(reg, val) };

    output(out, cmd, sizeof(cmd));

    return TRUE;
}


static boolean_t parseStopZe(FILE *out, const char *line)
{
    char      c;
    uint8_t reg;

    if(sscanf(line, "%c", &c) != 1 ||
       !getRegisterVal(c, &reg))
    {
        return FALSE;
    }

    uint8_t cmd[] = { M_SeqCmdStopRegZe(reg) };

    output(out, cmd, sizeof(cmd));

    return TRUE;
}


static boolean_t parseJumpZe(FILE *out, const char *line)
{
    char    c;
    int8_t  j;
    uint8_t reg;

    /* Skip leading space */
    while(isspace(*line))
    {
        line++;
        if(*line == '\0') return FALSE;
    }

    if(!isalpha(*line))
    {
        fprintf(stderr, "Error: Expected a character\n");
        return FALSE;
    }

    if(sscanf(line, "%c", &c) != 1 ||
       !getRegisterVal(c, &reg))
    {
        return FALSE;
    }

    /* Skip the remainder */
    while(!isspace(*line))
    {
        line++;
        if(*line == '\0') return FALSE;
    }

    while(isspace(*line))
    {
        line++;
        if(*line == '\0') return FALSE;
    }

    /* Try to parse as a signed number */
    if(sscanf(line, "%" SCNi8, &j) != 1)
    {
        uint32_t labelIdx;
        char    *name;

        /* Check for a label */
        if(*line != ':')
        {
            return FALSE;
        }

        /* Get and cleanup the name */
        name = trimdup(&line[1]);

        if(!getLabelIdx(name, &labelIdx))
        {
            /* Check if this is the initial parse for byte accounting,
             *  or actually generating the byte stream.  If the first
             *  parse, forward references may not yet be known.
             */
            if(out != NULL)
            {
                fprintf(stderr, "Error: Unknown label '%s'\n", name);
                free(name);
                return FALSE;
            }
            else
            {
                j = 0;
            }
        }
        else
        {
            j = labelData[labelIdx].byteOffset - outBytes;
        }

        free(name);
    }


    uint8_t cmd[] = { M_SeqCmdJumpRegZe(reg, j) };

    output(out, cmd, sizeof(cmd));

    return TRUE;
}


static boolean_t parsePreDef(FILE *out, const char *line)
{
    uint8_t n;

    if(sscanf(line, "%" SCNu8, &n) != 1)
    {
        return FALSE;
    }

    if(n > 0x1f)
    {
        fprintf(stderr, "Error: Predefined command number out of range (0-31)\n");
        return FALSE;
    }

    uint8_t cmd[] = { M_SeqCmdStartPreDef(n) };

    output(out, cmd, sizeof(cmd));

    return TRUE;
}


static boolean_t parseDelayReg(FILE *out, const char *line)
{
    char       c;
    uint8_t  reg;

    if(sscanf(line, "%c", &c) != 1 || !getRegisterVal(c, &reg))
    {
        return FALSE;
    }

    uint8_t cmd[] = { M_SeqCmdDelayRegTicks(reg) };

    output(out, cmd, sizeof(cmd));

    return TRUE;
}


static boolean_t parseLoadRand(FILE *out, const char *line)
{
    char       c;
    uint8_t  reg, min, max;

    if(sscanf(line, "%c %" SCNu8 " %" SCNu8, &c, &min, &max) != 3 ||
       !getRegisterVal(c, &reg))
    {
        return FALSE;
    }

    uint8_t cmd[] = { M_SeqCmdRandReg(reg, min, max) };

    output(out, cmd, sizeof(cmd));

    return TRUE;
}

static boolean_t parseLine(FILE *out, const char *line)
{
    const char     *cmd[] = { "c ", "stop", "jump ", "delayms ", "delaysec ",
                              "ack ", "fetch ", "load ", "add ", "dec ",
                              "stopze ", "jumpze ", "predef ", "delayregticks ",
                              "loadrand "};
    const handler   hnd[] = { parseColour, parseStop, parseJump, parseDelayMs, parseDelaySec,
                              parseAck, parseFetch, parseLoad, parseAdd, parseDec,
                              parseStopZe, parseJumpZe, parsePreDef, parseDelayReg,
                              parseLoadRand};
    uint8_t     c;

    for(c = 0; c < sizeof(cmd) / sizeof(cmd[0]); c++)
    {
        uint16_t len = strlen(cmd[c]);

        if(strncasecmp(cmd[c], line, len) == 0)
        {
            return hnd[c](out, &line[len]);
        }
    }

    return FALSE;
}

/** Display program licence and exit.
 */
static void licence(void)
{
    printf(
        "Cmdforge, a command assembler for Mike's Ambient Device.\n"
        "Copyright (C) 2007 Michael C McTernan, Michael.McTernan.2001@cs.bris.ac.uk\n"
        "\n"
        "This program is free software: you can redistribute it and/or modify\n"
        "it under the terms of the GNU General Public License as published by\n"
        "the Free Software Foundation, either version 3 of the License, or\n"
        "at your option) any later version.\n"
        "\n"
        "This program is distributed in the hope that it will be useful,\n"
        "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
        "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
        "GNU General Public License for more details.\n"
        "\n"
        "You should have received a copy of the GNU General Public License\n"
        "along with this program.  If not, see <http://www.gnu.org/licenses/>.\n");
}


/** Display program usage and exit.
 */
static void usage(void)
{
    printf("Usage: cmdforge [-eeprom] <infile> [outfile]\n"
           "       cmdforge -l\n"
           "\n"
           "Where:\n"
           "  -eeprom   Create an eeprom image suitable for reprogramming device.\n"
           "  <infile>  The file from which to read the text commands.  This maybe\n"
           "             '-' to cause reading from stdin.\n"
           "  <outfile> Optional file to which the binary commands will be written.  This\n"
           "             may be '-', or omitted entirely to cause writing to stdout.\n"
           "\n"
           "Cmdforge, Copyright (C) 2007 Michael C McTernan,\n"
           "                             Michael.McTernan.2001@cs.bris.ac.uk\n"
           "Cmdforge comes with ABSOLUTELY NO WARRANTY.  This is free software, and\n"
           "you are welcome to redistribute it under certain conditions;\n"
           "type `cmdforge -l' for details.\n");
}

/**************************************************************************
 * Global Functions
 **************************************************************************/

int main(const int argc, const char *argv[])
{
    boolean_t   eepromFormat = FALSE;
    char        line[2048];
    const char *inFile, *outFile;
    FILE       *in,     *out;

    /* Check if the licence should be displayed */
    if(argc == 2 && strcmp(argv[1], "-l") == 0)
    {
        licence();
        return EXIT_FAILURE;
    }

    /* Check if making an eeprom image */
    if((argc == 3 || argc == 4) && strcmp(argv[1], "-eeprom") == 0)
    {
        eepromFormat = TRUE;
        inFile  = argv[2];
        outFile = (argc == 4) ? argv[3] : "-";

    }
    else if(argc == 2 || argc == 3)
    {
        inFile  = argv[1];
        outFile = (argc == 3) ? argv[2] : "-";
    }
    else
    {
        usage();
        return EXIT_FAILURE;
    }

    /* Try to open the output file */
    if(strcmp("-", outFile) == 0)
    {
        out = stdout;
    }
    else
    {
        out = fopen(outFile, "wb");
        if(out == NULL)
        {
            perror("Failed to open output file");
            return EXIT_FAILURE;
        }
    }

    /* Try to open the input file */
    if(strcmp("-", inFile) == 0)
    {
        /* Open a temp file */
        in = tmpfile();
        if(in == NULL)
        {
            perror("Failed to open temporary file:");
            return EXIT_FAILURE;
        }

        /* Copy from stdin to the temp file */
        do
        {
            char     buf[1024];
            uint32_t r, w;

            r = fread(buf, 1, sizeof(buf), stdin);
            w = 0;

            while(r != w)
            {
                w += fwrite(&buf[w], 1, r - w, in);
            }
        }
        while(!feof(stdin));

        /* Move to start of input file */
        if(fseek(in, 0, SEEK_SET) != 0)
        {
            perror(NULL);
            return EXIT_FAILURE;
        }
    }
    else
    {
        in = fopen(inFile, "r");
        if(in == NULL)
        {
            perror("Failed to open input file");
            return EXIT_FAILURE;
        }
    }

    /* Build array of byte offsets */
    while(readLine(in, line, sizeof(line)))
    {
        /* Increase line count */
        lineNum++;

        if(!isCommentLine(line) && !isBlankLine(line))
        {
            stripComments(line);

            /* Check if the line contains a label */
            if(isLabelLine(line))
            {
                char *name;

                if(labelCount >= sizeof(labelData) / sizeof(labelData[0]))
                {
                    fprintf(stderr, "Error: Attempt to define too many labels\n");
                    return EXIT_FAILURE;
                }

                /* Duplicate the label name */
                name = trimdup(&line[1]);
                if(strlen(name) < 1)
                {
                    fprintf(stderr, "Error: Invalid label:\n%s\n", line);
                    return EXIT_FAILURE;
                }

                /* Insert padding for misaligned predefined labels */
                if(strncmp(name, "pd", 2) == 0 && outBytes & 1)
                {
                    outBytes++;
                }

                /* Store the label details */
                labelData[labelCount].label      = name;
                labelData[labelCount].byteOffset = outBytes;
                labelCount++;
            }
            else if(!parseLine(NULL, line))
            {
                fprintf(stderr, "Error: Failed to parse line %" PRIu32 ":\n%s\n", lineNum, line);
                return EXIT_FAILURE;
            }
        }
    }

    /* Check if writing an EEPROM image */
    if(eepromFormat)
    {
        /* Check there is space for one more label */
        if(labelCount >= sizeof(labelData) / sizeof(labelData[0]))
        {
            fprintf(stderr, "Error: Attempt to define too many labels\n");
            return EXIT_FAILURE;
        }

        /* Define end marker label */
        labelData[labelCount].label      = trimdup("pdEOF");
        labelData[labelCount].byteOffset = outBytes;
        labelCount++;

        if(!writeEepromHeader(NULL))
        {
           return EXIT_FAILURE;
        }
        writeEepromHeader(out);
    }
    else
    {
        uint8_t byteCount = outBytes;

        if(outBytes > 128)
        {
            fprintf(stderr, "Error: Sequence exceeds maximum permissible length: %" PRIu32 " bytes required, 128 allowed\n", outBytes);
            return EXIT_FAILURE;
        }

        /* Write a byte giving length of command sequence */
        output(out, &byteCount,1);
    }

    /* Return to the start of the input file */
    lineNum = 0;
    outBytes = 0;
    fseek(in, 0, SEEK_SET);

    /* Start parsing */
    while(readLine(in, line, sizeof(line)))
    {
        /* Increase line count */
        lineNum++;

        stripComments(line);

        if(!isCommentLine(line) && !isBlankLine(line))
        {
            if(isLabelLine(line))
            {
                /* Insert padding for misaligned predefined labels */
                if(strncmp(line, ":pd", 2) == 0 && outBytes & 1)
                {
                    uint8_t c = 0;

                    output(out, &c, 1);
                }
            }
            else if(!parseLine(out, line))
            {
                fprintf(stderr, "Error: Failed to parse line %" PRIu32 ":\n%s\n", lineNum, line);
                return EXIT_FAILURE;
            }
        }
    }

    /* Close off the streams */
    fclose(out);
    fclose(in);

    return EXIT_SUCCESS;
}


/* END OF FILE */
