/***************************************************************************
 * heartbeat.c: System load monitor for Mike's Ambient Device.
 * Copyright (C) 2007 Michael C McTernan,
 *    Michael.McTernan.2001@cs.bris.ac.uk
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ***************************************************************************/

/**************************************************************************
 * Include Files
 **************************************************************************/

#include <sys/types.h>
#include <sys/socket.h>
#if !defined(__CYGWIN__)
#include <sys/sysinfo.h>
#endif
#include <netdb.h>
#include <netinet/in.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "seq.h"

/**************************************************************************
 * Manifest Constants
 **************************************************************************/

#define HB_UPDATE_RATE_SECS  10
#define HB_LOAD_AVG_IDX      0



/**************************************************************************
 * Type Definitions
 **************************************************************************/

typedef struct
{
    unsigned long long int usr, nice, sys, idle, iowait, irq, softirq, steal;

    double total;
}
HbCpuStats;

/**************************************************************************
 * Variables
 **************************************************************************/

/**************************************************************************
 * Macros
 **************************************************************************/

/**************************************************************************
 * Local Functions
 **************************************************************************/

#if !defined(__CYGWIN__)
/** Parse contents of /proc/stat into a structure.
 *
 * \param[in,out] stats  Pointer to structure to fill with CPU statistics.
 */
static void getCpuStats(HbCpuStats *const stats)
{
    FILE *f;

    /* Clear the structure in case of failure */
    memset(stats, 0, sizeof(HbCpuStats));

    /* Open the stat file */
    f = fopen("/proc/stat", "r");
    if(f == NULL)
    {
        perror("fopen(/proc/stat) failed");
    }

    /* Try to read the first line */
    if(fscanf(f, "cpu  %lld %lld %lld %lld %lld %lld %lld %lld",
              &stats->usr, &stats->nice, &stats->sys, &stats->idle,
              &stats->iowait, &stats->irq, &stats->softirq, &stats->steal) < 8)
    {
        perror("failed to parse /proc/stat");
    }

    /* Produce a grand total.
     *  This gives a measure of the total number of samples passed since
     *  system startup.
     */
    stats->total = stats->usr + stats->nice + stats->sys + stats->idle +
                   stats->iowait + stats->irq + stats->softirq + stats->steal;

    fclose(f);
}


/** Compare two jiffy counts and return the difference.
 * This compares two jiffy counts and computes the diference in such a way
 * that overflow of the counter is correctly handled.
 *
 * \param[in] t1  The older value.
 * \param[in] t2  The newer value.
 */
static unsigned long long interval(unsigned long long t1,
                                   unsigned long long t2)
{
    if(t1 <= t2)
    {
        return t2 - t1;
    }
    else
    {
        return (((unsigned long long)-1) - t1) + t2 + 1;
    }
}


/** Compare two sets of stats and return the total interval.
 * This attempts to return the total passed jiffies between two sets
 * of counters.  Clearly it is possible for the result to overflow,
 * although this is only likely if the two sets have been taken a
 * long distance apart as the overflowing of individual counters
 * is handled through interval().
 *
 * \param[in] t1  The older result set.
 * \param[in] t2  The newer result set.
 */
static unsigned long long totalInterval(const HbCpuStats *t1,
                                        const HbCpuStats *t2)
{
    return interval(t1->usr, t2->usr) +
           interval(t1->nice, t2->nice) +
           interval(t1->sys, t2->sys) +
           interval(t1->idle, t2->idle) +
           interval(t1->iowait, t2->iowait) +
           interval(t1->irq, t2->irq) +
           interval(t1->softirq, t2->softirq) +
           interval(t1->steal, t2->steal);
}
#endif


/** Convert double precision fraction into an 8-bit unsigned value.
 * This takes a value in the range of 0 to 1 and converts it to an
 * 8-bit unsigned value in the range of 0 to 255.
 */
static uint8_t toColValue(double v)
{
    if(v >= 1.0)
    {
        return 255;
    }
    else if(v <= 0)
    {
        return 0;
    }
    else
    {
        return 255 * v;
    }
}


/** Build and send a command to the device.
 *
 * \param[in] sock    The socket on which to send the packet to the device.
 * \param[in] colour  Value in the range of 0-255 to determine the selected
 *                     output colour.
 * \param[in] speed   Value in the range of 0-255 to determine the pulse
 *                     speed.
 */
static void sendCmd(int sock, uint8_t colour, uint8_t speed)
{
    uint16_t delay;
    uint8_t  r, g, b, loops;


    if(colour <= 127)
    {
        r = colour * 2;
        g = 255;
    }
    else
    {
        r = 255;
        g = 256 - ((colour - 127) * 2);
    }
    b = 0;

    /* Compute pulse rate */
    delay = 7 * (255 - speed)  + 200;

    /* Compute how many loops are needed before the next update */
    loops = (HB_UPDATE_RATE_SECS * 1000) /
            (150+150+200+300+delay);


    /* Double count to allow for one missed message */

    loops *= 2;

    printf("%d: %d %d %d, delay=%u, loops=%u\n", colour, r, g, b, delay, loops);

    /* Construct the message */
    uint8_t msg[] = {
        0,
        M_SeqCmdRegSet(0,loops),
        M_SeqCmdRgb(0, 0, 0, 0),
        M_SeqCmdRgb(150, r, g, b),
        M_SeqCmdRgb(150, 0, 0, 0),
        M_SeqCmdRgb(200, r / 2, g / 2 , b /2),
        M_SeqCmdRgb(300, 0, 0, 0),
        M_SeqCmdDecReg(0, 1),
        M_SeqCmdDelayMs(delay),
        M_SeqCmdJumpRegZe(0, 4),
        M_SeqCmdJump(-21),
        M_SeqCmdRgb(2000, 0, 0,  10),  /* Lost contact sequence */
        M_SeqCmdRgb(2000, 0, 0,  20),
        M_SeqCmdRgb(2000, 0, 0,  30),
        M_SeqCmdRgb(2000, 0, 0,  40),
        M_SeqCmdRgb(2000, 0, 0,  50),
        M_SeqCmdRgb(2000, 0, 0,  60),
        M_SeqCmdRgb(2000, 0, 0,  70),
        M_SeqCmdRgb(2000, 0, 0,  80),
        M_SeqCmdRgb(2000, 0, 0, 127),
        M_SeqCmdRgb(2000, 0, 0, 255),
        M_SeqCmdStop()
    };

    /* Set the command sequence length */
    msg[0] = sizeof(msg) - 1;

    /* Send the message */
    send(sock, msg, sizeof(msg), 0);

}

/** Display program licence and exit.
 */
static void licence(void)
{
    printf(
        "Heartbeat, a system load monitor for Mike's Ambient Device.\n"
        "Copyright (C) 2007 Michael C McTernan, Michael.McTernan.2001@cs.bris.ac.uk\n"
        "\n"
        "This program is free software: you can redistribute it and/or modify\n"
        "it under the terms of the GNU General Public License as published by\n"
        "the Free Software Foundation, either version 3 of the License, or\n"
        "at your option) any later version.\n"
        "\n"
        "This program is distributed in the hope that it will be useful,\n"
        "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
        "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
        "GNU General Public License for more details.\n"
        "\n"
        "You should have received a copy of the GNU General Public License\n"
        "along with this program.  If not, see <http://www.gnu.org/licenses/>.\n");
}


/** Display program usage and exit.
 */
static void usage(void)
{
    printf("Usage: heartbeat <host> <colour_load> <speed_load>\n");
#if !defined(__CYGWIN__)
    printf("       heartbeat <host>");
#endif
    printf("Where:\n"
           "  <host>         Is the hostname or IP address of the device to\n"
           "                  control.\n"
           "  <colour_load>  Loading value that affects device colour, as a\n"
           "                  fractional value from 0-1.\n"
           "  <speed_load>   Loading value that affects beat speed, as a\n"
           "                  fractional value from 0-1.\n"
           "\n"
           "When a colour load and speed are specified, the device is updated\n"
           "to output the corresponding colour and beat pattern for around 10\n"
           "seconds.  The program must be re-ran before this time expires to\n"
           "update the beat pattern, otherwise the device will slowly turn blue.\n"
#if !defined(__CYGWIN__)
           "\n"
           "If ran without any load values, the program will monitor system loading\n"
           "itself and update the device every 10 seconds.\n"
#endif
           "\n"
           "Heartbeat, Copyright (C) 2007 Michael C McTernan,\n"
           "                              Michael.McTernan.2001@cs.bris.ac.uk\n"
           "Heartbeat comes with ABSOLUTELY NO WARRANTY.  This is free software, and\n"
           "you are welcome to redistribute it under certain conditions;\n"
           "type `heartbeat -l' for details.\n");
}

/**************************************************************************
 * Global Functions
 **************************************************************************/

int main(int argc, char *argv[])
{
    struct sockaddr_in a;
    struct hostent    *h;
    int                s;

    if(argc == 2 && strcmp(argv[1], "-l") == 0)
    {
        licence();
        return EXIT_FAILURE;
    }
    else if(argc != 4 && argc != 2)
    {
        usage();
        return EXIT_FAILURE;
    }

    /* Try to lookup the host */
    h = gethostbyname(argv[1]);
    if(h == NULL)
    {
        perror("gethostbyname() failed");
        return EXIT_FAILURE;
    }

    s = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if(s == -1)
    {
       perror("socket() failed");
       return EXIT_FAILURE;
    }

    /* Setup the address */
    a.sin_family     = AF_INET;
    a.sin_port       = htons(8888);
    memcpy(&a.sin_addr.s_addr, h->h_addr_list[0], 4);

    /* Connect to the host */
    if(connect(s, (struct sockaddr *)&a, sizeof(struct sockaddr_in)) == -1)
    {
        perror("connect() failed");
    }

    /* Check mode of operation */
    if(argc == 4)
    {
        sendCmd(s,
                toColValue(atof(argv[2])),
                toColValue(atof(argv[3])));
    }
    else
    {
#if !defined(__CYGWIN__)
        HbCpuStats now, prev;

        /* Prime the current stats */
        getCpuStats(&now);

        while(1)
        {
            int     nCpu = get_nprocs ();
            double  loadavg[3];
            double  taskLoad, ioLoad;

            /* Roll cpu stats and get more */
            prev = now;
            getCpuStats(&now);

            /* Get the CPU load average */
            getloadavg (loadavg, 3);

            /* Get load values */
            taskLoad = loadavg[HB_LOAD_AVG_IDX] / nCpu;
            ioLoad   = (double)interval(prev.iowait, now.iowait) /
                       (double)totalInterval(&prev, &now);

            sendCmd(s, toColValue(taskLoad), toColValue(ioLoad));

            sleep(HB_UPDATE_RATE_SECS);
        }
#endif
    }

    return EXIT_SUCCESS;
}

/* END OF FILE */
